@@ -18,6 +18,8 @@ from rdkit.Chem import AllChem
 import urllib
 from foldifyai.utils import get_type, file_exists
 import requests
+import fsspec
+
 try: 
     from logmd import LogMD
 except:
@@ -76,6 +78,20 @@ def compute_3d_conformer(mol, version: str = "v3") -> bool:
     return False
 
 def test(seq, affinity=False):
+    from pathlib import Path
+    import hashlib
+
+    cache_dir = Path.home() / ".cache"
+    cache_dir.mkdir(parents=True, exist_ok=True)
+    h = hashlib.sha256(seq.encode()).hexdigest()
+    f = cache_dir / h
+    if f.exists(): 
+        print(f'hit ligand test cache {seq}')
+        s = f.read_text()
+        return "ok" in s, s  # cache mechanism 
+    else: 
+        print(f"didn't find ligand test cache {seq}, testing...")
+
     try:
         mol = AllChem.MolFromSmiles(seq)
         mol = AllChem.AddHs(mol)
@@ -89,147 +105,198 @@ def test(seq, affinity=False):
                     f"{seq} has an atom with a name longer than "
                     f"4 characters: {atom_name}."
                 )
-                raise ValueError(msg)
-                return False 
+                (cache_dir / h).write_text("fail\n" + msg)
+                #raise ValueError(msg)
+                return False, msg
             atom.SetProp("name", atom_name)
 
         success = compute_3d_conformer(mol)
         if not success:
             msg = f"Failed to compute 3D conformer for {seq}"
-            return False 
-            raise ValueError(msg)
+            (cache_dir / h).write_text("fail\n" + msg)
+            return False, msg
 
         mol_no_h = AllChem.RemoveHs(mol, sanitize=False)
         affinity_mw = AllChem.Descriptors.MolWt(mol_no_h) if affinity else None
-        return True
+        (cache_dir / h).write_text("ok")
+        return True, ""
     except Exception as e:
         print(e, seq)
-        return False 
+        (cache_dir / h).write_text("fail\n" + str(e))
+        return False, str(e)
+
+def wait_on_server(host):
+    d = {0: '-', 1: '/', 2: '\\', 3: '|', 4: '.'}
+    i = 0 
+    t0 = time.time()
+    while True:
+        url = f"{host}/ping"
+        try:
+            text = urllib.request.urlopen(url).read().decode()
+            j = json.loads(text)
+            if j['alive']: return
+        except:
+            t = time.time()-t0
+            print(f'\rWaiting on server: {t:.2f}s {d[i%len(d)]}', end='')
+            i+=1
+            time.sleep(0.2)
+
 
 def fold(args):
     folder = args.input
     log = args.logmd 
+
+    if args.output == '': output_dir = f"foldify_{folder.replace('/','')}"
+    else: output_dir = args.output
+
+    s3_files = []
+    if args.s3 != '':
+        import boto3
+        s3 = boto3.session.Session(profile_name="r2").client('s3', endpoint_url=args.s3)
+        resp = s3.list_objects_v2(Bucket='dmitrij', Prefix=output_dir)
+        for obj in resp.get('Contents', []): 
+            s3_files.append(obj['Key'].replace('.zip','.fasta').replace('foldify_', ''))
+
+        if args.input[:5] != "s3://": 
+            num_local = len(files)
+            num_s3 = len(s3_files)
+
+            _files = []
+            num_overlap = 0
+            for file in files: 
+                if file not in s3_files: 
+                    _files.append(file)
+                else: 
+                    num_overlap += 1
+            files = _files
+
+            print(f"Local finished: {num_local}")
+            print(f"S3 finished: {num_s3} ")
+            print(f"Overlap: {num_overlap}. ")
+            print(f"Left: {len(files)}. ")
+
+        import fsspec
+        fs = fsspec.filesystem(
+            "s3",
+            profile="r2",
+            client_kwargs={
+                "endpoint_url": args.s3,
+                "use_ssl": True,
+            },
+            config_kwargs={
+                "s3": {"addressing_style": "path"} 
+            }
+        )
+        print("s3")
+
+    else: 
+        print("local")
+        fs = fsspec.filesystem("file")
+
     if args.input.endswith('.fasta'):
         files = [args.input]
         folder = folder.replace('.fasta', '')
     else:
-        files = [a for a in Path(folder).rglob("*.fasta")]
-        files = sorted(files, key=lambda p: os.path.getsize(str(p)))
-        files = [str(a) for a in files]
+        # [ ] port to s3 compatible. 
+        #files = [a for a in Path(folder).rglob("*.fasta")]
+        #files = sorted(files, key=lambda p: os.path.getsize(str(p)))
+        #files = [str(a) for a in files]
+        path = args.input[5:]
+        print(path)
+        files = fs.glob(f"{path}*.fasta")
+        files = sorted(files, key=lambda p: int(fs.size(p)))
+        path = path.split('/')
+        path = path[:1] + ['foldify_' + path[1]] + path[2:]
 
-    if log: 
-        l = LogMD()
+        path = "/".join(path)
 
-    if args.output == '': output_dir = f"foldify_{folder.replace('/','')}"
-    else: output_dir = args.output
-    print(files[:5])
-    #exit()
+        done = fs.glob(f"{path}*.zip")
+
+        zip_ids = {x.split('/')[-1].rsplit('.', 1)[0] for x in done}
+        left = [f"s3://{f}" for f in files if f.split('/')[-1].rsplit('.', 1)[0] not in zip_ids]
+
+      
+        print("S3 jobs: ", len(files))
+        print("S3 done: ", len(done))
+        print("Left: ", len(left))
+
+        files = left
+
+        # if s3 don't look for local files.
+        # but we want to look for stuff that's already finished. 
+
+    if log: l = LogMD()
+
+
+
+    wait_on_server(args.host)
 
     pbar = tqdm(files[::-1])
     for c,p in enumerate(pbar):
-        path = f"{output_dir}/{p.replace('.fasta', '')}/" #v0
-        #path = f"{output_dir}/" # v1
-        #print(path)
 
-        if args.cf != '':
-            s3_path = path[:-1] + '.zip'
-            s3_path = '/'.join(s3_path.split('/')[0::2])
-            #print(s3_path)
-            import boto3
-            # https://8a6ab2cee54f34a71f5a8d99e92da2d2.r2.cloudflarestorage.com
-            s3 = boto3.session.Session(profile_name="r2").client('s3', endpoint_url=args.cf)
-            if file_exists(s3, 'dmitrij', s3_path) and not args.override:
-                print(f"Skipping {p} found on s3 {s3_path}")
-                continue 
+        # 
+        if args.s3 != '': # handle single file vs folder. 
+            pass 
+            '''if "s3://" not in args.input: 
+                s3_path = path[:-1] + '.zip'
+                s3_path = '/'.join(s3_path.split('/')[0::2])
+            else:
+                s3_path = path[:-1] + '.zip'
+                s3_path = '/'.join(s3_path.split('/')[0::2])'''
+            s3_path = path + p.split('/')[-1].replace('.fasta', '.zip')
+        else: 
+            path = f"{output_dir}/{p.replace('.fasta', '')}/" 
+        #print(path, s3_path)
+        #exit()
 
-        #print(f"{path}boltz2_prediction_0.pdb")
-        if os.path.exists(f"{path}boltz2_prediction_0.pdb") and not args.override:
-            print(f"Skipping {p} found {path}boltz2_prediction_0.pdb")
+        # [ ] s3 compatible. 
+        if False: #os.path.exists(f"{path}boltz2_prediction_0.pdb") and not args.override:
+            print(f"Local: Skipping {p} found {path}boltz2_prediction_0.pdb")
 
-            # if args.cf != '' it doesn't exist on s3 but does exist locally => upload it. 
-            if args.cf != '' and not file_exists(s3, 'dmitrij', s3_path):
+            # if args.s3 != '' it doesn't exist on s3 but does exist locally => upload it. 
+            # this may have killed the data i already computed. 
+            if args.s3 != '' and not file_exists(s3, 'dmitrij', s3_path):
                 print(f"Uploading {p} to s3 {s3_path}")
                 zip_bytes = io.BytesIO()
-                #print(os.listdir(path))
-                #print(s3_path)
                 with zipfile.ZipFile(zip_bytes, 'w') as zip_ref:
-                    #for file in os.listdir(path):
-                    #    zip_ref.write(os.path.join(path, file), arcname=file)
                     for root, dirs, files in os.walk(path):
                         for file in files:
                             zip_ref.write(os.path.join(root, file), os.path.relpath(os.path.join(root, file), path))
                 zip_bytes.seek(0)
-                #print(zip_bytes)
                 with open('my_folder.zip', 'wb') as f: 
                    f.write(zip_bytes.getvalue())
                 zip_bytes.seek(0)
-
-                a = s3.upload_fileobj(zip_bytes, 'dmitrij', s3_path)
-                #print(a)
-                #exit()
-
+                s3.upload_fileobj(zip_bytes, 'dmitrij', s3_path)
             continue 
-        #print('stopping')
-        #continue # exit()
+
         try: 
-            t0 = time.time()
             skip = False 
+
             p = str(p)
+            #content = open(p).read()
+            content = fs.open(p, "rt").read()
 
-            # change this to point to output_dir
-            #new_path = p.replace('.fasta','_raw.pdb')
-            #if os.path.exists(new_path): continue 
-            #if os.path.exists(path): continue 
-
-            content = open(p).read()
             num_tokens = sum([len(line) for line in content.split('\n') if not line.startswith('>')])
 
-
-            # load paired MSA send seperate. 
-            if False: 
-                lines = [line for line in content.split('\n') if len(line) > 1]
-                headers, seqs = lines[::2], lines[1::2]
-                a3m_paths = [header.split('|')[-1] for header in headers]
-                a3m_paths = [path for path in a3m_paths if os.path.exists(path)]
-                if len(a3m_paths) == len(headers):
-                    # assuming all the same 
-                    assert a3m_paths[0] == a3m_paths[-1], f"Assuming paired MSA, but found {a3m_paths[0]} != {a3m_paths[1]}."
-                    msa = open(f"{a3m_paths[0]}", 'r').read()
-                    msas = msa.split('\x00')
-                    for seq in seqs: 
-                        for msa in msas:
-                            if seq in msa: 
-                                # pair
-                                pass
-
-            #print(num_tokens)
-            #if num_tokens > 1000: continue 
-
-            if True:  # this killed my desktop
-                for line in content.split('\n'):
-                    if line.startswith('>'): continue 
-                    if line == '': continue 
-                    if get_type(line) == 'SMILES': 
-                        if not test(line): 
-                            print(f"Skipping {p}. RDKit didn't like {line}. ")
-                            #open(new_path, 'w').write(f"Skipping {p}. RDKit didn't like {line}. ")
-                            skip = True 
-                        else: print('ok')
-                if skip: continue 
+            for line in content.split('\n'):
+                if line.startswith('>'): continue 
+                if line == '': continue 
+                if get_type(line) == 'SMILES': 
+                    if not test(line): 
+                        print(f"Skipping {p}. RDKit didn't like {line}. ")
+                        #open(new_path, 'w').write(f"Skipping {p}. RDKit didn't like {line}. ")
+                        skip = True 
+                    else: print('ok')
+            if skip: continue 
             encoded = urllib.parse.quote(content, safe="")
             if len(content) == 0: 
                 continue 
 
-            # decode the .a3m file and pass this on. 
-
-            url = f"{args.host}/fold?only_return_zip=True&seq={encoded}&{args.args}&gpu={args.gpu}&get_msa_from_server={args.msa}"
+            url = f"{args.host}/fold?ui=False&only_return_zip=True&seq={encoded}&{args.args}&gpu={args.gpu}&get_msa_from_server={args.msa}"
 
             # Open connection with progress reporting
             response = urllib.request.urlopen(url)
-            total_size = int(response.headers.get('content-length', 0))
             block_size = 1024
-            #progress_bar = tqdm(total=total_size, unit='iB', unit_scale=True, 
-            #                    desc=f"Foldify")
             
             result = ''
             while True:
@@ -237,10 +304,8 @@ def fold(args):
                 if not data:
                     break
                 result += data.decode('utf-8')
-                #pbar.update(len(data))
                 pbar.set_description(f"{Fore.BLUE}{Style.BRIGHT}[Foldify]{Style.RESET_ALL} {time.strftime('%d-%m %H:%M:%S')} {p} tokens={num_tokens} {len(result)/1000}KB")
 
-            #with open(new_path, 'w') as f: f.write(result)
             jsons = [json.loads(a) for a in result.split('\n@\n') if a != '']
             # last is a zip file, unzip. 
             b64_zip_data = jsons[-1]['data']
@@ -248,15 +313,24 @@ def fold(args):
             zip_bytes = base64.b64decode(b64_zip_data)
             zip_in_memory = io.BytesIO(zip_bytes)
             with zipfile.ZipFile(zip_in_memory, 'r') as zip_ref:
-                os.makedirs(path, exist_ok=True)
-                zip_ref.extractall(path)
-                if args.cf != '':
-                    import boto3
+                # [ ] remove s3 from this path 
+                # this path didn't change in each iteration? 
+                #os.makedirs(path, exist_ok=True)
+                #zip_ref.extractall(path) 
+
+                #if args.s3 != '' and args.input[:5] != "s3://":
+                with fs.open(f"{s3_path}", 'wb') as f:
+                    f.write(zip_bytes)
+                #    # integrate s3 with logmd -- parallel download s3 zips and store localcache? 
+                '''elif args.s3 != '':
+
+                    #import boto3
                     # https://8a6ab2cee54f34a71f5a8d99e92da2d2.r2.cloudflarestorage.com
-                    s3 = boto3.session.Session(profile_name="r2").client('s3', endpoint_url=args.cf)
+                    #s3 = boto3.session.Session(profile_name="r2").client('s3', endpoint_url=args.s3)
                     #s3.upload_file('foldify_folders.zip', 'dmitrij', 'foldify_folders.zip')
                     zip_in_memory.seek(0)
                     s3.upload_fileobj(zip_in_memory, 'dmitrij', s3_path) # 'foldify_folders.zip')
+                    print(s3_path)'''
 
             #time.sleep(1)
 
@@ -266,6 +340,7 @@ def fold(args):
             pass 
         #time.sleep(1)
         print('')
+        wait_on_server(args.host)
         #exit()
 
 
@@ -282,9 +357,10 @@ def main() -> None:  # pragma: no cover
     parser.add_argument('-output', '-o', type=str, default='', help='Output directory for results')
     parser.add_argument('-gpu', '-g', type=int, default=0, help='GPU')
     parser.add_argument('-y', action='store_true', help='Pre-accept using remote host. ')
-    parser.add_argument('-cf', type=str, default='', help='Cloudflared endpoint url to store online.')
+    parser.add_argument('-s3', type=str, default='', help='S3 endpoint (developed for cloudflared to lower cost).')
     parser.add_argument('-override', action='store_true', help='Override existing files')
     parser.add_argument('-msa', type=str, default='', help='Get msa from other ip. ')
+
     args = parser.parse_args()
 
     if args.host == 'https://gpu1.foldify.org' and not args.y:
